/*
 * ll_hsusart_impl.h
 *
 *  Created on: Oct 29, 2019
 *      Author: abody
 */
#include <f4ll/packet_usart.h>
#include <string.h>

namespace f4ll {

template <typename T> static inline T round_up_to_4(T input)
{
    return (input + 3) & (((T)-1) - 3);
}

packet_usart::packet_usart(USART_TypeDef *usart, DMA_TypeDef *dma, uint32_t streamRx, uint32_t streamTx)
    : usart_core(usart, dma, streamRx, streamTx)
{
    crc_handler::instance().attach_slot(m_crcSlot);
    LL_USART_EnableIT_IDLE(usart);
    LL_USART_EnableIT_ERROR(usart);
}

void packet_usart::RxProcessed(bool second)
{
    m_rxBuffers[second].busy = false;
    m_rxBuffers[second].error = false;
}

void packet_usart::SetCallback(IHsUsartCallback *callback, uintptr_t callbackParam)
{
    m_userCallback = callback;
    m_userCallbackParam = callbackParam;
}

void packet_usart::PostPacket(uint8_t const *payload, uint8_t length, bool wait_for_crc_queue)
{
    uint16_t payload_length = round_up_to_4((uint16_t)length);

    BuildHeader(m_txBuffer.packet, m_txSerialNo++, length);
    if (payload) {
        memcpy(m_txBuffer.packet.payload, payload, length);
    }
    m_txBuffer.requestedLength = sizeof(m_txBuffer.packet.header) + payload_length + sizeof(uint32_t);
    m_txBuffer.busy = true;
    m_txBuffer.error = false;

    crc_handler::instance().enqueue(
        m_crcSlot, 0, &m_txBuffer.packet, sizeof(packet_header) + payload_length, nullptr,
        reinterpret_cast<uintptr_t>(m_txBuffer.packet.payload + payload_length));

    while (wait_for_crc_queue && crc_handler::instance().is_queued(m_crcSlot, 0))
        ;

    setup_transmit(&m_txBuffer.packet, m_txBuffer.requestedLength);

    ++m_stats.sent;
}

void packet_usart::SetupReceive()
{
    m_rxBuffers[m_rxBufferSelector].requestedLength = sizeof(m_rxBuffers[m_rxBufferSelector].packet);
    usart_core::setup_receive(&m_rxBuffers[m_rxBufferSelector], sizeof(m_rxBuffers[m_rxBufferSelector].packet));
}

//////////////////////////////////////
// UsartCore pure virtual functions //
//////////////////////////////////////

void packet_usart::receiver_idle(void)
{
    uint16_t rcvdLen = m_rxBuffers[m_rxBufferSelector].requestedLength - LL_DMA_GetDataLength(m_rx_dma.get_dma(), m_rx_dma.get_stream());

    if (rcvdLen >= sizeof(packet_header)) {
        if (CheckHeader(m_rxBuffers[m_rxBufferSelector].packet.header)) {
            if (rcvdLen >= sizeof(packet_header) + round_up_to_4((uint16_t)m_rxBuffers[m_rxBufferSelector].packet.header.payload_length) +
                               sizeof(uint32_t)) {
                LL_DMA_DisableStream(m_rx_dma.get_dma(), m_rx_dma.get_stream());
            } else {
                ++m_stats.premature_payload;
            }
        } else {
            m_rxBuffers[m_rxBufferSelector].error = 1;
            LL_DMA_DisableStream(m_rx_dma.get_dma(), m_rx_dma.get_stream());
        }
    } else {
        ++m_stats.premature_hdr;
    }
}

void packet_usart::transmission_complete(void)
{
    LL_USART_DisableDirectionTx(m_usart); // enforcing an idle frame
    LL_USART_EnableDirectionTx(m_usart);
    m_txBuffer.busy = 0;
}

void packet_usart::framing_error(void) {}

void packet_usart::overrun(void) {}

void packet_usart::rx_dma_transfer_complete(void)
{
    if (CheckHeader(m_rxBuffers[m_rxBufferSelector].packet.header)) {
        crc_handler::instance().enqueue(
            m_crcSlot, 1, &m_rxBuffers[m_rxBufferSelector].packet,
            sizeof(packet_header) + round_up_to_4((uint16_t)m_rxBuffers[m_rxBufferSelector].packet.header.payload_length), this,
            m_rxBufferSelector);
    } else {
        ++m_stats.hdrError;
        m_rxBuffers[m_rxBufferSelector].error = true;
    }
    SwitchRxBuffers();
}

void packet_usart::rx_dma_half_transfer(void) {}

void packet_usart::rx_dma_error(dma_helper::dma_error_type reason)
{
    (void)reason;

    m_rxBuffers[m_rxBufferSelector].error = 1;
    ++m_stats.rxDmaError;
    SwitchRxBuffers();
}

void packet_usart::tx_dma_transfer_complete(void)
{
    LL_USART_EnableIT_TC(m_usart);
    LL_DMA_DisableStream(m_tx_dma.get_dma(), m_tx_dma.get_stream());
}

void packet_usart::tx_dma_half_transfer(void) {}

void packet_usart::tx_dma_error(dma_helper::dma_error_type reason)
{
    (void)reason;

    m_txBuffer.error = 1;
    ++m_stats.txDmaError;
}

///////////////////////
// Private functions //
///////////////////////

void packet_usart::BuildHeader(Packet &packet, uint8_t serialNo, uint8_t length)
{
    uint8_t hash = STARTMARKER;

    packet.header.start_byte = STARTMARKER;
    packet.header.serial = serialNo;
    hash ^= serialNo;
    packet.header.payload_length = length;
    hash ^= length;
    packet.header.hash = hash;
}

bool packet_usart::CheckHeader(packet_header &header)
{
    return header.start_byte == STARTMARKER && (header.start_byte ^ header.serial ^ header.payload_length) == header.hash;
}

void packet_usart::SwitchRxBuffers(void)
{
    ++m_stats.rcvd;
    m_rxBufferSelector = !m_rxBufferSelector;

    if (m_rxBuffers[m_rxBufferSelector].busy) {
        ++m_stats.overrun;
    }
    SetupReceive();
}

///////////////////////////
// crc_handler::ICallback //
///////////////////////////

void packet_usart::crc_succeeded(uintptr_t callbackParam, uint32_t crc, uint8_t task)
{
    (void)task;

    Buffer &buf(m_rxBuffers[static_cast<int>(callbackParam)]);

    buf.busy = 1;
    if (*(uint32_t *)(buf.packet.payload + round_up_to_4((uint16_t)buf.packet.header.payload_length)) != crc) {
        buf.error = 1;
        buf.errorInfo = crc;
        ++m_stats.payloadErrror;
    }
    if (m_userCallback) {
        buf.busy = !m_userCallback->PacketReceived(this, m_userCallbackParam, buf.packet);
    }
}

void packet_usart::crc_failed(uintptr_t callbackParam, uint32_t crc, uint8_t task)
{
    (void)crc;
    (void)task;

    Buffer &buf(m_rxBuffers[static_cast<int>(callbackParam)]);
    buf.busy = buf.error = true;
    buf.errorInfo = 0;
    ++m_stats.payloadErrror;
    if (m_userCallback) {
        buf.busy = !m_userCallback->PacketReceived(this, m_userCallbackParam, buf.packet);
    }
}

} //	namespace f4ll
